
# ModelChecker.py
#  Implementacin del algoritmo Model Checking
#

from Tkinter import *
import debug
from ASG import *
from ATOM3 import *
from string import split
from ATOM3TypeDialog import *
from ATOM3Text import *

class ModelChecker :

	#----------------------------------------------------------------------
	#
	# Cada estado del grafo contiene las variables:
	# . listFormulas . lista de frmulas que se cumplen en el estado
	#
	# Cada nodo del rbol sintctico de la frmula contiene las variables:
	# . subformula   . expresin de la subfrmula a partir de dicho nodo.
	#                  En el nodo raiz toma como valor la frmula completa.
	#

	#----------------------------------------------------------------------
	#
	# Constructor
	#
	def __init__(self) :

		self.TRUE  = 1	# constante para el valor true
		self.FALSE = 0	# constante para el valor false
		self.DEBUG = 0	# constante para DEBUG

	#----------------------------------------------------------------------
	# METODOS PARA MANEJAR LA LISTA DE FORMULAS DE UN ESTADO
	#----------------------------------------------------------------------
	# 
	# Devuelve true si un estado est etiquetado con la frmula f, y false
	# en caso contrario.
	#
	def labeled (self, state, formula):

		return formula in state.listFormulas

	#----------------------------------------------------------------------
	#
	# Aade una frmula a la lista de frmulas que se cumplen en un estado.
	#
	def add_label (self, state, formula) :

		state.listFormulas.append(formula)

	#----------------------------------------------------------------------
	# METODOS PARA VERIFICAR FORMULAS
	#----------------------------------------------------------------------
	# 
	# Etiqueta los nodos del grafo donde se cumple la frmula NOT(f).
	#
	# PARAMETROS:
	# - self
	# - graph: grafo de exploracin
	# - f: operando de la frmula NOT(f)
	#
	# RETORNO: la expresin.
	#
	def NOT (self, graph, f) :

		# Construimos la expresin a evaluar
		NOT_f = "NOT(" + f + ")"

		# Etiquetamos los nodos que no cumplen f
		for node in graph.listNodes.get('RGnode'):
			if not self.labeled(node, f): 
				self.add_label(node, NOT_f)

		# Devolvemos la expresin
		return NOT_f

	#----------------------------------------------------------------------
	# 
	# Etiqueta los nodos del grafo donde se cumple la frmula (f1 AND f2).
	#
	# PARAMETROS:
	# - self
	# - graph: grafo de exploracin
	# - f1: primer operando de la frmula (f1 AND f2)
	# - f2: segundo operando de la frmula (f1 AND f2)
	#
	# RETORNO: la expresin.
	#
	def AND (self, graph, f1, f2) :

		# Construimos la expresin a evaluar
		f1_AND_f2 = "(" + f1 + " AND " + f2 + ")"

		# Etiquetamos los nodos que cumplen f1 y f2
		for node in graph.listNodes.get('RGnode'):
			if self.labeled(node, f1) and self.labeled(node, f2): 
				self.add_label(node, f1_AND_f2)

		# Devolvemos al expresin
		return f1_AND_f2

	#----------------------------------------------------------------------
	# 
	# Etiqueta los nodos del grafo donde se cumple la frmula AX(f).
	# PROCEDIMIENTO: etiquetamos con la frmula AX(f) todos aquellos nodos 
	# que tengan TODOS sus descendientes etiquetados con f.
	#
	# PARAMETROS:
	# - self
	# - graph: grafo de exploracin
	# - f: operando de la frmula AX(f)
	#
	# RETORNO: la expresin.
	#
	def AX (self, graph, f) :

		# Construimos la expresin a evaluar
		AX_f = "AX(" + f + ")"

		# Para cada nodo comprobamos si todos sus descendientes cumplen f
		for node in graph.listNodes.get('RGnode'):
			if not self.labeled(node, AX_f):
				FLAG = None
				for child in node.out_connections_:
					FLAG = self.labeled(child.out_connections_[0], f)
					if not FLAG: break
				# slo si todos los hijos cumplen f etiquetamos el nodo
				if FLAG: self.add_label(node, AX_f)

		# Devolvemos la expresin
		return AX_f
				

	#----------------------------------------------------------------------
	# 
	# Etiqueta los nodos del grafo donde se cumple la frmula EX(f).
	# PROCEDIMIENTO: etiquetamos con la frmula EX(f) todos aquellos nodos 
	# que tengan ALGN descendientes etiquetado con f.
	#
	# PARAMETROS:
	# - self
	# - graph: grafo de exploracin
	# - f: operando de la frmula EX(f)
	#
	# RETORNO: la expresin.
	#
	def EX (self, graph, f) :

		# Construimos la expresin a evaluar
		EX_f = "EX(" + f + ")"

		# Para cada nodo comprobamos si algn descendiente cumple f
		for node in graph.listNodes.get('RGnode'):
			if not self.labeled(node, EX_f):
				for child in node.out_connections_:
					# si un hijo cumple f, etiquetamos el nodo y salimos
					if self.labeled(child.out_connections_[0], f):
						self.add_label(node, EX_f)
						break

		# Devolvemos la expresin
		return EX_f

	#----------------------------------------------------------------------
	# 
	# Etiqueta los nodos del grafo donde se cumple la frmula A(f1 U f2).
	# PROCEDIMIENTO: se utiliza una bsqueda en profundidad primero para 
	# explorar el grafo. El procedimiento recursivo AU_recursive realiza 
	# la comprobacin de la frmula AU en el node que recibe, devolviendo
	# true o false segn cumpla o no la frmula.
	#
	# PARAMETROS:
	# - self
	# - graph: grafo de exploracin
	# - f1: primer operando de la frmula A(f1 U f2)
	# - f2: segundo operando de la frmula A(f1 U f2)
	#
	# RETORNO: la expresin.
	#
	def AU (self, graph, f1, f2) :

		# Construimos la expresin a evaluar
		A_f1_U_f2 = "A(" + f1 + " U " + f2 + ")"

		# Marcamos todos los estados a false (no han sido visitados por el algoritmo).
		# Cuando el nodo sea visitado por el algoritmo, esa variable se pondr a TRUE.
		for node in graph.listNodes.get('RGnode'): node.marked = self.FALSE

		# Para cada nodo comprobamos si cumple la frmula A(f1 U f2)
		for node in graph.listNodes.get('RGnode'):
			if not node.marked: self.AU_recursive(node, f1, f2)

		# Devolvemos la expresin
		return A_f1_U_f2

	# --- RUTINA RECURSIVA ---
	
	def AU_recursive (self, node, f1, f2) :

		# Construimos la expresin a evaluar
		A_f1_U_f2 = "A(" + f1 + " U " + f2 + ")"

		# Si el nodo ya se ha visitado y est etiquetado con la formula 
		# devolvemos TRUE; pero si se ha visitado y no est etiquetado
		# con la formula, devolvemos FALSE.
		if node.marked: return self.labeled(node, A_f1_U_f2)

		# Marcamos el nodo a comprobar como visitado. 
		node.marked = self.TRUE

		# 1) A(f1 U f2) se cumple en el nodo si f2 se cumple en el nodo. En ese
		#    caso etiquetamos el nodo con la frmula y devolvemos TRUE.
		# 2) A(f1 U f2) no se cumple en el nodo si f1 no se cumple en el nodo.
		#    En ese caso devolvemos FALSE.		
		if self.labeled(node, f2):
			self.add_label(node, A_f1_U_f2)
			return self.TRUE
		elif not self.labeled(node, f1):
			return self.FALSE

		# Si hemos llegado aqu sabemos que en el nodo se cumple f1 pero no f2.
		# Ahora hay que comprobar si A(f1 U f2) es cierta para todos los sucesores 
		# del nodo. Si encontramos algn sucesor que no cumple la frmula, entonces 
		# la formula tampoco se cumple en el nodo. Pero si la formula es cierta en
		# todos los sucesores, entonces tambin es cierta en el nodo.
		for child in node.out_connections_:		
			if not self.AU_recursive(child.out_connections_[0], f1, f2):
				return self.FALSE
		self.add_label(node, A_f1_U_f2)
		return self.TRUE

	#----------------------------------------------------------------------
	# 
	# Etiqueta los nodos del grafo donde se cumple la frmula E(f1 U f2).
	# PROCEDIMIENTO: encontramos todos los estados que estn etiquetados 
	# con f2. Usando la conversin inversa de las transiciones en el grafo, 
	# etiquetamos todos los estados que pueden ser alcanzados por un path 
	# en el cual cada estado est etiquetado con f1.
	#
	# PARAMETROS:
	# - self
	# - graph: grafo de exploracin
	# - f1: primer operando de la frmula E(f1 U f2)
	# - f2: segundo operando de la frmula E(f1 U f2)
	#
	# RETORNO: la expresin.
	#
	def EU (self, graph, f1, f2) :

		# Construimos la expresin a evaluar
		E_f1_U_f2 = "E(" + f1 + " U " + f2 + ")"

		# Construimos una lista con todos los nodos etiquetados con f2
		listNodesToCheck = [node for node in graph.listNodes.get('RGnode') if self.labeled(node, f2)]

		# Todos los nodos que cumplen f2 cumplen, a su vez, E_f1_U_f2
		for node in listNodesToCheck: self.add_label(node, E_f1_U_f2)

		# Mientras haya nodos que verificar en la lista
		while len(listNodesToCheck)>0:
			node = listNodesToCheck.pop(0)						#Procesamos el siguiente nodo de la lista.
			for parent in node.in_connections_:						#Comprobamos cada predecesor directo del nodo. Si el nodo
				if not(self.labeled(parent.in_connections_[0], E_f1_U_f2)) \
				   and self.labeled(parent.in_connections_[0], f1):		#an no est etiquetado con E_f1_U_f2 pero cumple f1,
					self.add_label(parent.in_connections_[0], E_f1_U_f2)	#lo etiquetamos
					listNodesToCheck.append(parent.in_connections_[0])	#y comprobamos sus padres
					
		# Devolvemos la expresin
		return E_f1_U_f2

	#----------------------------------------------------------------------
	# RUTINA PRINCIPAL: Implementacin del algoritmo Model Checking
	#----------------------------------------------------------------------
	# Comprueba si la frmula recibida se cumple para el grafo dado
	#
	# PARAMETROS:
	# - self
	# - graph: grafo de exploracin
	# - formula: frmula a verificar
	#
	# RETORNO: la expresin.
	#
	def check (self, graph, formula) :

		# 1. Inicializamos la lista de formulas que se cumplen en cada estado
		#    (inicialmente son los predicados que se dan en su variable state).
		#    Los guardamos como una lista para manejarlos mejor.
		#    Casos especiales:
		#    1.1. Aadimos el predicado "True" a cada estado para que labeled(estado,"True") devuelva TRUE.
		#    1.2. Si un estado no tiene salidas lo etiquetamos con el predicado especial "deadlock".
		
		for node in graph.listNodes.get('RGnode'):
			listFormulas = node.state.getValue()              
			listFormulas = listFormulas[1:len(listFormulas)-1] # borramos []
			node.listFormulas = [strip(f) for f in split(listFormulas,',')]
			self.add_label(node,"True")                                       # <- caso especial 1.1
			if len(node.out_connections_)==0: self.add_label(node,"deadlock") # <- caso especial 1.2
			if self.DEBUG: print node.id.getValue(), node.listFormulas

		# 2. Procedemos a etiquetar cada estado con las formulas que cumple.
		#    Para ello vamos recorriendo el rbol sintctico con la frmula,
		#    desde las hojas hasta la raiz, comprobando si cada subfrmula 
		#    se cumple en los estados. Si una frmula se cumple en un estado,
		#    ser aadida a la lista de frmulas del estado.

		# 2.1. Obtenemos el nodo raiz (el que no tiene entradas)

		root = None
		for node in formula.listNodes.get('CTLUnary')  + \
				formula.listNodes.get('CTLBinary') + \
				formula.listNodes.get('CTLPredicate'):
			if len(node.in_connections_)==0:
				root = node
				break

		# 2.2. Creamos la lista listStages, que contendr una sublista por cada
		#      nivel de profundidad del rbol sintctico. Cada sublista contendr
		#      los nodos (operandos, operadores) que pertenecen al nivel. Ejemplo:
		#
		#      [ [...], 				<- nivel 2 de profundidad (descendientes de desc_root y desc2_root)
		#        [desc1_root, desc2_root],  <- nivel 1 de profundidad (descendientes de root)
		#        [root] ]				<- nivel 0 de profundidad (nodo raiz)


		i_stage    = [root]    # en el primer nivel slo esta la raiz
		listStages = [i_stage] # creamos una lista de niveles
		while len(i_stage)>0:
			i_stage = []
			for node in listStages[0]: # aadimos al nivel i+1 los descendientes del nivel i
				i_stage = i_stage + [child.out_connections_[0] for child in node.out_connections_]
			if len(i_stage)>0: listStages.insert(0, i_stage)
		if self.DEBUG: print listStages

		# 2.3. Recorremos los niveles desde el ms profundo al ms exterior,
		#      obteniendo las subfrmulas desde la ms interna a la completa.
		#      Para facilitar la tarea guardaremos en cada nodo la subfrmula
		#      que se construye a partir del mismo.

		for i_stage in listStages:
			for node in i_stage:

				# PREDICATE (nodos HOJA -> ya estn evaluados en el grafo) -------------------

				if node.getClass()=="CTLPredicate":
					node.subformula = node.value.getValue()

				# BINARY OPERATOR (AND/AU/EU) ------------------------------------------------

				elif node.getClass()=="CTLBinary":
					# Obtenemos el operador y los dos operandos
					operator = node.value.getValue()[1]
					if node.out_connections_[0].getClass()=="CTLBinaryOperLeft":
						operand1 = node.out_connections_[0].out_connections_[0].subformula
						operand2 = node.out_connections_[1].out_connections_[0].subformula
					else:
						operand1 = node.out_connections_[1].out_connections_[0].subformula
						operand2 = node.out_connections_[0].out_connections_[0].subformula
					# Llamamos al mtodo adecuado para el tipo de operador
					if   operator == 0: node.subformula = self.AND(graph, operand1, operand2)
					elif operator == 1: node.subformula = self.AU (graph, operand1, operand2)
					elif operator == 2: node.subformula = self.EU (graph, operand1, operand2)

				# UNARY OPERATOR (NOT/AX/EX) -------------------------------------------------

				elif node.getClass()=="CTLUnary":
					# Obtenemos el operador y el operando
					operator = node.value.getValue()[1]
					operand  = node.out_connections_[0].out_connections_[0].subformula
					# Llamamos al mtodo adecuado para el tipo de operador
					if   operator == 0: node.subformula = self.NOT(graph, operand)
					elif operator == 1: node.subformula = self.AX (graph, operand)
					elif operator == 2: node.subformula = self.EX (graph, operand)

				if self.DEBUG: print node.subformula
			
		# 2.4. Imprimimos el id de los nodos que cumplen la frmula.

		results = [node.id.getValue() for node in graph.listNodes.get('RGnode') if self.labeled(node, root.subformula)]
		res_data = "ETIQUETAS DE LOS NODOS\n----------------------\n"
		for node in graph.listNodes.get('RGnode'):
			res_data+=str(node.id.getValue())+str(node.listFormulas)+"\n"
		res_data+= "  --> FORMULA = "+str(root.subformula)+"\n"
		res_data+= "  --> LISTA DE NODOS QUE LA VERIFICAN = "+str(results)		
		resText = ATOM3Text(res_data, 110, 30)
		at3d = ATOM3TypeDialog(formula.parent, resText)
				
	#----------------------------------------------------------------------
